#include "stdafx.h"
#include "processed_impl.h"
#include "dump_helpers.h"
#include "hexdump.h"
#include "unmangle.h"
#include <sstream>

std::string dbobj_pe::get_name() const 
{ 
	std::stringstream ss;
	ss << "PE";
	IMAGE_FILE_HEADER * hdr = (IMAGE_FILE_HEADER *)m_base;

	unsigned long chars = hdr->Characteristics;
	if (chars & IMAGE_FILE_DLL) 
		ss << " DLL";
	else if (chars & IMAGE_FILE_EXECUTABLE_IMAGE) 
		ss << " EXE";
	return ss.str();
}

std::string dbobj_anonymous_object::get_name() const 
{ 
	return std::string("ANONYMOUS OBJECT");
}

std::string dbobj_object::get_name() const 
{ 
	return std::string("OBJECT");
}

std::string dbobj_coff_section::get_name() const
{
	std::stringstream ss;
	ss << m_id << ": " << m_sh.name();
	return ss.str();
}

std::string image_characteristics(unsigned long mask)
{
	std::string pad("                   ");
	std::stringstream ss;

	unsigned long i = 1;
	for (int x=0;x<16;++x, i<<=1)
	{
		unsigned long mm = mask & i;
		switch (mm)
		{
		case IMAGE_FILE_RELOCS_STRIPPED: ss << pad << "Relocations stripped\n"; break;
		case IMAGE_FILE_EXECUTABLE_IMAGE: ss << pad << "Executable\n"; break;
		case IMAGE_FILE_LINE_NUMS_STRIPPED: ss << pad << "Line numbers stripped\n"; break;
		case IMAGE_FILE_LOCAL_SYMS_STRIPPED: ss << pad << "Symbols stripped\n"; break;
		case IMAGE_FILE_AGGRESIVE_WS_TRIM: ss << pad << "Aggressively trim working set\n"; break; 
		case IMAGE_FILE_LARGE_ADDRESS_AWARE: ss << pad << "Application can handle large (>2GB) addresses\n"; break; 
		case IMAGE_FILE_BYTES_REVERSED_LO: ss << pad << "Bytes reversed\n"; break;   
		case IMAGE_FILE_32BIT_MACHINE: ss << pad << "32 bit word machine\n"; break;       
		case IMAGE_FILE_DEBUG_STRIPPED: ss << pad << "Debug information stripped\n"; break;
		case IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP: ss << pad << "CD - run from swapfile\n"; break;   
		case IMAGE_FILE_NET_RUN_FROM_SWAP: ss << pad << "Net - run from swapfile\n"; break; 
		case IMAGE_FILE_SYSTEM: ss << pad << "System File\n"; break;
		case IMAGE_FILE_DLL: ss << pad << "DLL\n"; break;           
		case IMAGE_FILE_UP_SYSTEM_ONLY: ss << pad << "Uniprocessor Only\n"; break;
		case IMAGE_FILE_BYTES_REVERSED_HI: 
//			ss << pad << "Big Endian\n"; 
			break; 
		}
	}
	return ss.str();
}

std::string dbobj_coff_header::get_display() 
{ 
	IMAGE_FILE_HEADER & h = *m_hdr;

	std::stringstream ss;
	ss << "FILE HEADER VALUES\n";
	ss << lpad(hex(h.Machine), 16)				<< " machine (" << machine_type(h.Machine) << ")\n";
	ss << lpad(hex(h.NumberOfSections), 16)		<< " number of sections\n";
	ss << lpad(hex(h.TimeDateStamp), 16)			<< " time date stamp" << stamptotime(h.TimeDateStamp) << "\n";
	ss << lpad(hex(h.PointerToSymbolTable), 16)	<< " file pointer to symbol table\n";
	ss << lpad(hex(h.NumberOfSymbols), 16)		<< " number of symbols\n";
	ss << lpad(hex(h.SizeOfOptionalHeader), 16)	<< " size of optional header\n";
	ss << lpad(hex(h.Characteristics), 16)		<< " characteristics\n";
	// Decode characteristics
	ss << image_characteristics(h.Characteristics) << "\n";

	return ss.str();
}

std::string dbobj_import_directory::get_display()
{
	IMAGE_IMPORT_DESCRIPTOR * pii = (IMAGE_IMPORT_DESCRIPTOR*)this->get_start();
	cofpcdat & cd = this->get_cd();

	std::stringstream ss;

	// We know we're in the same section, get a better helper for rva deciphering.
	image_section_helper sh = cd.get_rva_section(pii->OriginalFirstThunk ? pii->OriginalFirstThunk : pii->FirstThunk);
	ss << "  Section contains the following imports:\n\n";
	do
	{

		ss << "    " << (char*)sh.get_rva(pii->Name) << "\n";
		ss << "              " << lpad(hex(cd.imagebase + pii->FirstThunk), 8)      << " Import Address Table\n";
		ss << "              " << lpad(hex(pii->OriginalFirstThunk ? (cd.imagebase + pii->OriginalFirstThunk) : 0), 8) << " Import Name Table\n";

		ss << "              " << lpad(hex(pii->TimeDateStamp), 8)		 << " time date stamp" << stamptotime(pii->TimeDateStamp) << "\n";
		
		ss << "              " << lpad(hex(pii->ForwarderChain), 8)		 << " Index of first forwarder reference\n";
		ss << "\n";
		// halmacpi.dll has IATA in a different section from the local one!
		DWORD * pp = pii->OriginalFirstThunk ? (DWORD*)cd.get_rva(pii->OriginalFirstThunk) : NULL;
		DWORD * pn = (DWORD*)cd.get_rva(pii->FirstThunk);
		bool bound = (pii->TimeDateStamp != 0);
		if (!pp) pp = pn;
		if (pp && *pp) // Just in case!
		{
			do 
			{
				ss << "      " << ( !bound ? "        " : lpad(hex(*pn),8,'0') ) << " "; 
				if (*pp & 0xF0000000) // top bit set - ordinal.
				{
					unsigned long ord = (*pp & ~0xF0000000);
					ss << "       Ordinal " << lpad(str(ord), 5) << "\n";
				}
				else
				{
					IMAGE_IMPORT_BY_NAME * iin = (IMAGE_IMPORT_BY_NAME*)sh.get_rva(*pp);
					ss << lpad(hex(iin->Hint), 6) << " " << iin->Name << "\n";
				}
				pp++;
				pn++;
			} while (*pp);
		}
		ss << "\n";
	} while ((++pii)->OriginalFirstThunk || pii->FirstThunk);
	return ss.str();
}

std::string debug_type(int type)
{
	char * stype[] = {
		"(   0)",
		"coff",
		"cv",
		"fpo",
		"misc",
		"except",
		"fixup",
		"ots",
		"ofs",
		"borl",
		"(   A)",
		"clsid",
	};
	return stype[type];
}

typedef struct CV_HEADER_ 
{
	typedef enum
	{
		CV_NB10 = '01BN',
		CV_RSDS = 'SDSR',
	};
	DWORD CvSignature; // NBxx
} CV_HEADER;

typedef struct CV_NB10_ 
{
	CV_HEADER  Header; 
	LONG  Offset;      // Always 0 for NB10
	DWORD      Signature;     
	DWORD      Age;           
	char       Name[1];
} CV_NB10;

typedef struct CV_RSDS_ 
{
	CV_HEADER  Header; 
	GUID       g;
	DWORD      x;
	char       Name[1];
} CV_RSDS;

std::string dump_debug_misc_data(byte * start, int len)
{
	IMAGE_DEBUG_MISC * dbg = (IMAGE_DEBUG_MISC*)start;
	if (!dbg->Unicode)
	{
		std::string ret = (char*)(dbg->Data);
		return ret;
	};
	return "";
}

// CodeView dump.
std::string dump_debug_cv_data(byte * start, int len)
{
	std::stringstream ss;

	CV_HEADER * cvh = (CV_HEADER *)start;

	ss << "Format: ";
	switch (cvh->CvSignature)
	{
	case CV_HEADER::CV_NB10:
		{
			CV_NB10 * nb10h = (CV_NB10*)start;
			ss << "NB10, " << hex(nb10h->Signature) << ", " << nb10h->Age << ", " << nb10h->Name;
		}
		break;
	case CV_HEADER::CV_RSDS:
		{
			USES_CONVERSION;
			CV_RSDS * rsdsh = (CV_RSDS*)start;
			OLECHAR tmp[128];

			StringFromGUID2(rsdsh->g, tmp, sizeof(tmp));
			ss << "RSDS, " << OLE2A(tmp) << ", " << rsdsh->x << ", " << rsdsh->Name << "\n";
		}
		break;
	default:
		ss << std::string((char*)start, 4);
	}
	return ss.str();
}

std::string dbobj_debug_directory::get_display()
{
	IMAGE_DEBUG_DIRECTORY * pdd = (IMAGE_DEBUG_DIRECTORY*)this->get_start();
	cofpcdat & cd = this->get_cd();
	int size = this->get_size();

	std::stringstream ssc;

	ssc << "  Debug Directories\n\n";
	ssc << "        Time Type       Size      RVA  Pointer\n";
	ssc << "    -------- ------ -------- -------- --------\n";

	int ndif = size / sizeof(IMAGE_DEBUG_DIRECTORY);
	for (int i=0;i<ndif;++i)
	{
		IMAGE_DEBUG_DIRECTORY & dd = pdd[i];
		ssc << "    " << lpad(hex(dd.TimeDateStamp), 8, '0') << " ";
		ssc << rpad(debug_type(dd.Type), 6) << " ";
		ssc << lpad(hex(dd.SizeOfData), 8) << " ";
		ssc << lpad(hex(dd.AddressOfRawData), 8, '0') << " ";
		ssc << lpad(hex(dd.PointerToRawData), 8);
		// msvcp70 is an example of a dll which uses an out of section data block.
		byte * pdata = dd.AddressOfRawData ? cd.get_rva(dd.AddressOfRawData) : (cd.base + dd.PointerToRawData);
		switch (dd.Type)
		{
		case IMAGE_DEBUG_TYPE_CODEVIEW:
			ssc  << "    " << dump_debug_cv_data(pdata, dd.SizeOfData) << "\n";
			break;
		case IMAGE_DEBUG_TYPE_RESERVED10:
			if (dd.AddressOfRawData || dd.PointerToRawData)
			{
				ssc  << "    " << lpad(hex(*(DWORD*)pdata), 8, '0') << "\n";
			}
			else
			{
				ssc << "\n";
			}
			break;
		case IMAGE_DEBUG_TYPE_MISC:
			ssc  << "    Image Name: " << dump_debug_misc_data(pdata, dd.SizeOfData) << "\n";
			break;
		default:
			ssc << "\n";
			break;
		}
	}
	ssc << "\n";

	return ssc.str();
}


typedef struct DELAYED_IMPORT_ENTRY_T_
{
	DWORD Characteristics;
	DWORD AddressOfName;
	DWORD AddressOfHMODULE;
	DWORD ImportAddressTable;
	DWORD ImportNameTable;
	DWORD BoundImportNameTable;
	DWORD UnboundImportNameTable;
	DWORD TimeDateStamp;
	void adjustup(DWORD imagebase)
	{
		if (AddressOfHMODULE) AddressOfHMODULE += imagebase;
		if (ImportAddressTable) ImportAddressTable += imagebase;
		if (ImportNameTable) ImportNameTable += imagebase;
		if (BoundImportNameTable) BoundImportNameTable += imagebase;
		if (UnboundImportNameTable) UnboundImportNameTable += imagebase;
	}
	void adjustdown(DWORD imagebase)
	{
		if (AddressOfHMODULE) AddressOfHMODULE -= imagebase;
		if (ImportAddressTable) ImportAddressTable -= imagebase;
		if (ImportNameTable) ImportNameTable -= imagebase;
		if (BoundImportNameTable) BoundImportNameTable -= imagebase;
		if (UnboundImportNameTable) UnboundImportNameTable -= imagebase;
	}
} DELAYED_IMPORT_ENTRY;


std::string dbobj_delay_import_directory::get_display()
{
	std::stringstream ssc;
	cofpcdat & cd = this->get_cd();
	image_section_helper & sh = m_sh;

	ssc << "  Section contains the following delay load imports:\n\n";

	DELAYED_IMPORT_ENTRY * de = (DELAYED_IMPORT_ENTRY *)this->get_start();
	while (de->AddressOfHMODULE)
	{
		bool adjustdown = false;
		DELAYED_IMPORT_ENTRY dec = *de;
		if (de->Characteristics & 0x01)
		{
			dec.adjustup(cd.imagebase);
		}
		else
		{
			adjustdown = true;
		}
		// no guarantee it's in the same section - eg acrotray.exe
		ssc << "    " << cd.get_rva((de->Characteristics == 0) ? (dec.AddressOfName - cd.imagebase) : dec.AddressOfName) << "\n";
		ssc << "              " << lpad(hex(dec.Characteristics), 8, '0') << " Characteristics\n";
		ssc << "              " << lpad(hex(dec.AddressOfHMODULE), 8, '0') << " Address of HMODULE\n";
		ssc << "              " << lpad(hex(dec.ImportAddressTable), 8, '0') << " Import Address Table\n";
		ssc << "              " << lpad(hex(dec.ImportNameTable), 8, '0') << " Import Name Table\n";
		ssc << "              " << lpad(hex(dec.BoundImportNameTable), 8, '0') << " Bound Import Name Table\n";
		ssc << "              " << lpad(hex(dec.UnboundImportNameTable), 8, '0') << " Unload Import Name Table\n";
		ssc << "              " << lpad(hex(dec.TimeDateStamp), 8, dec.TimeDateStamp ? '0' : ' ') << " time date stamp" << stamptotime(dec.TimeDateStamp);
		ssc << "\n\n";
		dec.adjustdown(cd.imagebase);

		DWORD * iat = (DWORD*)cd.get_rva(dec.ImportAddressTable);
		DWORD * pnm = (DWORD*)cd.get_rva(dec.ImportNameTable);
		for (;*pnm; ++iat, ++pnm)
		{
			if (*pnm & 0x80000000) // top bit set - ordinal.
			{
				unsigned long ord = (*pnm & ~0xF0000000);
				ssc << "          " << lpad(hex(*iat), 8, '0') << "                 Ordinal " << lpad(str(ord), 5) << "\n";
			}
			else
			{
				unsigned long addr = *pnm;
				if (adjustdown) addr -= cd.imagebase;
				IMAGE_IMPORT_BY_NAME * pbn = (IMAGE_IMPORT_BY_NAME*)sh.get_rva(addr);
				ssc << "          " << lpad(hex(*iat), 8, '0') << lpad(str(pbn->Hint), 16) << " " << pbn->Name << "\n"; 
			}
		}
		ssc << "\n";

		++de;
	}
	return ssc.str();
}

std::string str_dll_characteristics(unsigned long chrs)
{
	if (!chrs) return "";

	std::stringstream ssc;

	unsigned long bit = 0x1;
	for (;bit <= 0x8000;bit<<=1)
	{
		switch (bit & chrs)
		{
		case 0x1: case 0x2: case 0x4: case 0x8:
//      IMAGE_LIBRARY_PROCESS_INIT           0x0001     // Reserved.
//      IMAGE_LIBRARY_PROCESS_TERM           0x0002     // Reserved.
//      IMAGE_LIBRARY_THREAD_INIT            0x0004     // Reserved.
//      IMAGE_LIBRARY_THREAD_TERM            0x0008     // Reserved.
			ssc << "                   RESERVED - UNKNOWN\n";
			break;			
		case 0x40: // IMAGE_DLLCHARACTERISTICS_DYNAMIC_BASE: <--- not defined in vs2005
			ssc << "                   Dynamic base\n";
			break;
		case 0x100:
			ssc << "                   NX compatible\n";
			break;
		case IMAGE_DLLCHARACTERISTICS_NO_ISOLATION:
			ssc << "                   Don't want isolation\n";
			break;
		case IMAGE_DLLCHARACTERISTICS_NO_SEH:
			ssc << "                   No structured exception handler\n";
			break;
		case IMAGE_DLLCHARACTERISTICS_NO_BIND:
			ssc << "                   Do not bind this image\n";
			break;
		case IMAGE_DLLCHARACTERISTICS_WDM_DRIVER:
			ssc << "                   WDM Driver\n";
			break;
		case IMAGE_DLLCHARACTERISTICS_TERMINAL_SERVER_AWARE:
			ssc << "                   Terminal Server Aware\n";
			break;
		default:
			break;
		}
	}

	return ssc.str();
}

std::string nt_subsystem(int x)
{
	switch (x)
	{
	default:
	case IMAGE_SUBSYSTEM_UNKNOWN: return "Unknown";
	case IMAGE_SUBSYSTEM_NATIVE: return "Native";
	case IMAGE_SUBSYSTEM_WINDOWS_GUI: return "Windows GUI";
	case IMAGE_SUBSYSTEM_WINDOWS_CUI: return "Windows CUI";
	case IMAGE_SUBSYSTEM_OS2_CUI: return "OS/2 CUI";
	case IMAGE_SUBSYSTEM_POSIX_CUI: return "Posix CUI";
	case IMAGE_SUBSYSTEM_NATIVE_WINDOWS: return "Win9x";
	case IMAGE_SUBSYSTEM_WINDOWS_CE_GUI: return "Windows CE GUI";
	case IMAGE_SUBSYSTEM_EFI_APPLICATION: return "EFI Application";
	case IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER: return "EFI Boot Service Driver";
	case IMAGE_SUBSYSTEM_EFI_RUNTIME_DRIVER: return "EFI Runtime Driver";
	case IMAGE_SUBSYSTEM_EFI_ROM: return "EFI Rom";
	case IMAGE_SUBSYSTEM_XBOX: return "Xbox";
	}
}


std::string dbobj_coff_opt_header::get_display()
{
	std::stringstream ssc;

	IMAGE_OPTIONAL_HEADER32 * hdr = m_hdr;
	cofpcdat & cd = *m_pcd;

	ssc << lpad(hex(hdr->Magic), 16)	<< " magic # (PE32)\n";
	ssc << lpad(majmin(hdr->MajorLinkerVersion, hdr->MinorLinkerVersion), 16)		<< " linker version\n";
	ssc << lpad(hex(hdr->SizeOfCode), 16) << " size of code\n";
	ssc << lpad(hex(hdr->SizeOfInitializedData), 16) << " size of initialized data\n";
	ssc << lpad(hex(hdr->SizeOfUninitializedData), 16) << " size of uninitialized data\n";
	ssc << lpad(hex(hdr->AddressOfEntryPoint), 16) << " entry point";
	if (hdr->AddressOfEntryPoint) 
	{
		ssc << " (" << lpad(hex(hdr->ImageBase + hdr->AddressOfEntryPoint), 8, '0') << ")";
		std::string alias;
		if (cd.test_symtable_alias(hdr->AddressOfEntryPoint, alias))
		{
			ssc << " " << alias;
		}
	}
	ssc << "\n";
	ssc << lpad(hex(hdr->BaseOfCode), 16) << " base of code\n";
	ssc << lpad(hex(hdr->BaseOfData), 16) << " base of data\n";
	ssc << lpad(lpad(hex(hdr->ImageBase), 8), 16) << " image base (" << lpad(hex(hdr->ImageBase), 8, '0') << " to " << lpad(hex(hdr->ImageBase + hdr->SizeOfImage - 1), 8, '0') << ")\n";
	ssc << lpad(hex(hdr->SectionAlignment), 16) << " section alignment\n";
	ssc << lpad(hex(hdr->FileAlignment), 16) << " file alignment\n";
	ssc << lpad(majmin(hdr->MajorOperatingSystemVersion, hdr->MinorOperatingSystemVersion), 16) << " operating system version\n";
	ssc << lpad(majmin(hdr->MajorImageVersion, hdr->MinorImageVersion), 16) << " image version\n";
	ssc << lpad(majmin(hdr->MajorSubsystemVersion, hdr->MinorSubsystemVersion), 16) << " subsystem version\n";
	ssc << lpad(hex(hdr->Win32VersionValue), 16) << " Win32 version\n";
	ssc << lpad(hex(hdr->SizeOfImage), 16) << " size of image\n";
	ssc << lpad(hex(hdr->SizeOfHeaders), 16) << " size of headers\n";
	ssc << lpad(hex(hdr->CheckSum), 16) << " checksum\n";
	ssc << lpad(hex(hdr->Subsystem), 16) << " subsystem (" << nt_subsystem(hdr->Subsystem) << ")\n";
	ssc << lpad(hex(hdr->DllCharacteristics), 16) << " DLL characteristics\n";
	ssc << str_dll_characteristics(hdr->DllCharacteristics);
	ssc << lpad(hex(hdr->SizeOfStackReserve), 16) << " size of stack reserve\n";
	ssc << lpad(hex(hdr->SizeOfStackCommit), 16) << " size of stack commit\n";
	ssc << lpad(hex(hdr->SizeOfHeapReserve), 16) << " size of heap reserve\n";
	ssc << lpad(hex(hdr->SizeOfHeapCommit), 16) << " size of heap commit\n";
	ssc << lpad(hex(hdr->LoaderFlags), 16) << " loader flags\n";
	ssc << lpad(hex(hdr->NumberOfRvaAndSizes), 16) << " number of directories\n";

	return ssc.str();
}

std::string dbobj_export_directory::get_display()
{
	IMAGE_EXPORT_DIRECTORY * id = (IMAGE_EXPORT_DIRECTORY *)this->get_start();
	cofpcdat & cd = this->get_cd();

	image_section_helper sh = cd.get_rva_section(id->Name);

	std::stringstream ssc;

	ssc << "  Section contains the following exports for " << (char *)sh.get_rva(id->Name) << "\n\n";
	ssc << "    " << lpad(hex(id->Characteristics), 8, '0') << " characteristics\n";
	ssc << "    " << lpad(hex(id->TimeDateStamp),8) << " time date stamp" << stamptotime(id->TimeDateStamp, true) << "\n";
	ssc << "    " << lpad(majmin(id->MajorVersion, id->MinorVersion), 8)	<< " version\n";
	ssc << "    " << lpad(str(id->Base), 8)	<< " ordinal base\n";
	ssc << "    " << lpad(str(id->NumberOfFunctions), 8)	<< " number of functions\n";
	ssc << "    " << lpad(str(id->NumberOfNames), 8)	<< " number of names\n";
	ssc << "\n";
	ssc << "    ordinal hint RVA      name\n\n";

	long  * pfunctions = NULL;
	short * pordinals  = NULL;
	long  * pnames     = NULL;
	if (id->NumberOfFunctions)
	{
		pfunctions = (long*)sh.get_rva(id->AddressOfFunctions);
	}
	if (id->NumberOfNames)
	{
		pordinals  = (short*)sh.get_rva(id->AddressOfNameOrdinals);
		pnames     = (long*)sh.get_rva(id->AddressOfNames);
	}
	std::vector<char> vseen_ordinals;
	vseen_ordinals.resize(id->NumberOfFunctions);
	if (id->NumberOfNames && !pfunctions) throw std::runtime_error("Missing functions info in exports directory");
	for (unsigned int i=0;i<id->NumberOfNames;++i)
	{
		unsigned long maliasid = 0;
		ssc << lpad(str(pordinals[i] + id->Base), 11) << " ";
		vseen_ordinals[pordinals[i]] = true;
		ssc << lpad(hex(i), 4) << " ";

		char * nameptr = (char*)sh.get_rva(pnames[i]);
		maliasid = pfunctions[pordinals[i]];

		std::string dispname = (char*)sh.get_rva(pnames[i]);
		std::string unmandis = unmangle(dispname);

		if (cd.test_rva(pfunctions[pordinals[i]]) && ((pfunctions[pordinals[i]] - pnames[i]) == (dispname.length() + 1)))
		{
			ssc << "         " << dispname << " (forwarded to " << (char*)(sh.get_rva(pfunctions[pordinals[i]])) << ")";
		}
		else
		{
			ssc << lpad(hex(pfunctions[pordinals[i]]), 8, '0') << " " << dispname;
		}

#if 0
		if (unmandis != "")
		{
			ssc << " = " << dispname << " (" << unmandis << ")";
		}
#endif

		std::string alias;
		if (cd.test_symtable_alias(maliasid, alias))
		{
			ssc << " = " << alias;
		}
		ssc << "\n";
	}
	for (unsigned int i=0;i < id->NumberOfFunctions;++i)
	{
		if (pfunctions[i] && !vseen_ordinals[i])
		{
			ssc << lpad(str(i + id->Base), 11) << " ";
			ssc << "     ";
			ssc << lpad(hex(pfunctions[i]), 8, '0') << " [NONAME]\n";
		}
	}
	ssc << "\n";

	return ssc.str();
}

std::string dbobj_exception_directory::get_display()
{
	std::stringstream ssc;

	IMAGE_CE_RUNTIME_FUNCTION_ENTRY * fe = reinterpret_cast<IMAGE_CE_RUNTIME_FUNCTION_ENTRY *>(this->get_start());
	int count = this->get_size() / sizeof(IMAGE_CE_RUNTIME_FUNCTION_ENTRY);
	ssc << "Function Table (" << count << ")\n\n";
	ssc << "           Begin    End      Prolog   Excpt 32bit Fixup Function Name\n\n";

	for (int x=0;x<count;++x, ++fe)
	{
		
		ssc << "  " << lpad(hex(x*sizeof(IMAGE_CE_RUNTIME_FUNCTION_ENTRY)), 8, '0') <<
					 " " << lpad(hex(fe->FuncStart), 8, '0') <<  
					 " " << lpad(hex(fe->FuncStart + 4*fe->FuncLen), 8, '0') <<
					 " " << lpad(hex(fe->FuncStart + 4*fe->PrologLen), 8, '0') <<
					 "   " << (fe->ExceptionFlag ? "Y" : "N") <<  
					 "     " << (fe->ThirtyTwoBit ? "Y" : "N") << "     ";
		ssc << "N\n"; // THIS IS WRONG.  WHERE DOES THE FIXUP INFO COME FROM??
	}
	ssc << "\n";

	return ssc.str();
}

std::string dbobj_load_config_directory::get_display()
{
	std::stringstream ssc;

	IMAGE_LOAD_CONFIG_DIRECTORY32* ldcc = (IMAGE_LOAD_CONFIG_DIRECTORY32*)this->get_start();
	cofpcdat & cd = this->get_cd();

	ssc << "  Section contains the following load config:\n\n";

	if (ldcc->Size != sizeof(IMAGE_LOAD_CONFIG_DIRECTORY32)) throw std::runtime_error("unhandled loadconfig dd");

	ssc << "            " << lpad(hex(ldcc->Size), 8, '0') << " size\n";
	ssc << "            " << lpad(hex(ldcc->TimeDateStamp), 8) << " time date stamp\n";
	ssc << "            " << lpad(majmin(ldcc->MajorVersion, ldcc->MinorVersion), 8) << " Version\n";
	ssc << "            " << lpad(hex(ldcc->GlobalFlagsClear), 8) << " GlobalFlags Clear\n";
	ssc << "            " << lpad(hex(ldcc->GlobalFlagsSet), 8) << " GlobalFlags Set\n";
	ssc << "            " << lpad(hex(ldcc->CriticalSectionDefaultTimeout), 8) << " Critical Section Default Timeout\n";
	ssc << "            " << lpad(hex(ldcc->DeCommitFreeBlockThreshold), 8) << " Decommit Free Block Threshold\n";
	ssc << "            " << lpad(hex(ldcc->DeCommitTotalFreeThreshold), 8) << " Decommit Total Free Threshold\n";
	ssc << "            " << lpad(hex(ldcc->LockPrefixTable), 8, '0') << " Lock Prefix Table\n";
	ssc << "            " << lpad(hex(ldcc->MaximumAllocationSize), 8) << " Maximum Allocation Size\n";
	ssc << "            " << lpad(hex(ldcc->VirtualMemoryThreshold), 8) << " Virtual Memory Threshold\n";
	ssc << "            " << lpad(hex(ldcc->ProcessHeapFlags), 8) << " Process Heap Flags\n";
	ssc << "            " << lpad(hex(ldcc->ProcessAffinityMask), 8) << " Process Affinity Mask\n";
	ssc << "            " << lpad(hex(ldcc->CSDVersion), 8) << " CSD Version\n";
	ssc << "            " << lpad(lpad(hex(ldcc->Reserved1),4,'0'), 8) << " Reserved\n";
	ssc << "            " << lpad(hex(ldcc->EditList), 8, '0') << " Edit list\n";
	ssc << "            " << lpad(hex(ldcc->SecurityCookie), 8, '0') << " Security Cookie\n";
	ssc << "            " << lpad(hex(ldcc->SEHandlerTable), 8, '0') << " Safe Exception Handler Table\n";
	ssc << "            " << lpad(hex(ldcc->SEHandlerCount), 8) << " Safe Exception Handler Count\n";
	if (ldcc->SEHandlerCount > 0)
	{
		ssc << "\n";
		ssc << "    Safe Exception Handler Table\n\n";
		ssc << "          Address\n";
		ssc << "          --------\n";

		DWORD * eh = (DWORD*)cd.get_rva(ldcc->SEHandlerTable - cd.imagebase);
		for (unsigned int x=0;x<ldcc->SEHandlerCount;++x,++eh)
		{
			ssc << "          " << lpad(hex(cd.imagebase + *eh),8,'0') << "\n";
		}
	}
	ssc << "\n";

	return ssc.str();
}

std::string relocation_type(int type)
{
	switch (type)
	{
	case IMAGE_REL_I386_ABSOLUTE:
		return "ABSOLUTE";  
	case IMAGE_REL_I386_DIR16:
		return "DIR16";  
	case IMAGE_REL_I386_REL16:
		return "REL16";  
	case IMAGE_REL_I386_DIR32:
		return "DIR32";  
	case IMAGE_REL_I386_DIR32NB:
		return "DIR32NB";  
	case IMAGE_REL_I386_SEG12:
		return "SEG12";  
	case IMAGE_REL_I386_SECTION:
		return "SECTION";  
	case IMAGE_REL_I386_SECREL:
		return "SECREL";  
	case IMAGE_REL_I386_TOKEN:
		return "TOKEN";  
	case IMAGE_REL_I386_SECREL7:
		return "SECREL7";  
	case IMAGE_REL_I386_REL32:      
		return "REL32";  
	default:
		return "??????";	
	}
}

int relsize(int type)
{
	switch (type)
	{
	case IMAGE_REL_I386_ABSOLUTE:
	case IMAGE_REL_I386_TOKEN:
	case IMAGE_REL_I386_SECREL7: // ??
		return 0;
	case IMAGE_REL_I386_DIR16:
	case IMAGE_REL_I386_REL16:
	case IMAGE_REL_I386_SEG12: // Not supported,
		return 0;
	case IMAGE_REL_I386_DIR32:
	case IMAGE_REL_I386_DIR32NB:
	case IMAGE_REL_I386_SECREL:
	case IMAGE_REL_I386_REL32:      
		return 4;
	case IMAGE_REL_I386_SECTION:
		return 2;
	default:
		return 0;
	}
}


std::string dump_relocation(IMAGE_RELOCATION & r, byte * data, cofpcdat & cd)
{
	std::stringstream ssc;
	ssc << " " << lpad(hex(r.VirtualAddress), 8, '0') << "  ";
	ssc << rpad(relocation_type(r.Type), 16) << "  ";
	ssc << lpad(rawbytes(data + r.VirtualAddress, relsize(r.Type)), 17) << "  ";
	ssc << lpad(hex(r.SymbolTableIndex), 8) << "  ";
	std::string name = symbol_name(cd.psym[r.SymbolTableIndex], cd.stringtable);
	ssc << name;
	std::string umname = unmangle(name);
	if (umname != "") ssc << " (" << umname << ")";
	ssc << "\n";
	return ssc.str();
}


std::string dbobj_section_linenumbers::get_display()
{
	IMAGE_SECTION_HEADER & h = *(m_s->get_section_header());
	cofpcdat & cd = m_s->get_cd();

	std::stringstream ssc;

	ssc << "LINENUMBERS #" << hex(m_s->get_id()) << "\n";

	IMAGE_LINENUMBER * lnm = (IMAGE_LINENUMBER *)(cd.base + h.PointerToLinenumbers);
	int basenumber = 0;
	int offset = 0;
	for (int i=0; i<h.NumberOfLinenumbers; ++i,++lnm)
	{
		if (!lnm->Linenumber)
		{
			IMAGE_SYMBOL * psym = (cd.psym + lnm->Type.SymbolTableIndex);
			if (psym->StorageClass == IMAGE_SYM_CLASS_STATIC ||
				psym->StorageClass == IMAGE_SYM_CLASS_EXTERNAL)
			{
				IMAGE_AUX_SYMBOL * paux = (IMAGE_AUX_SYMBOL *)psym+1;
				psym = (cd.psym + paux->Sym.TagIndex);
			}
			IMAGE_AUX_SYMBOL * paux = (IMAGE_AUX_SYMBOL *)psym+1;
			basenumber = paux->Sym.Misc.LnSz.Linenumber;
			ssc << "\n Symbol index: " << lpad(hex(lnm->Type.SymbolTableIndex), 8) << " Base line number: " << lpad(str(basenumber), 5) << "\n";
			std::string name = symbol_name(*(cd.psym + lnm->Type.SymbolTableIndex), cd.stringtable);
			std::string umname = unmangle(name);
			ssc << " Symbol name = " << name;
			if (umname != "") ssc << " (" << umname << ")";
			ssc << "\n";
		}
		else
		{
			ssc << " " << (offset == 0 ? "" : " ") << lpad(hex(lnm->Type.VirtualAddress), 8, '0') << "(" << lpad(str(basenumber + lnm->Linenumber),5) << ")";
			if (++offset == 4) 
			{
				offset = 0;
				ssc << "\n";
			}
		}
	}
	ssc << "\n\n";
	return ssc.str();
}

std::string dbobj_section_relocations::get_display()
{
	IMAGE_SECTION_HEADER & h = *(m_s->get_section_header());
	cofpcdat & cd = m_s->get_cd();

	std::stringstream ssc;

	ssc << "RELOCATIONS #" << hex(m_s->get_id()) << "\n";
	ssc << "                                                Symbol    Symbol\n" <<
				 " Offset    Type              Applied To         Index     Name\n" << 
				 " --------  ----------------  -----------------  --------  ------\n";

	if (h.NumberOfRelocations == 0xffff)
	{
		ssc << "Extended relocations NYI\n";
	}
	else
	{
		IMAGE_RELOCATION * rel0 = (IMAGE_RELOCATION *)(cd.base + h.PointerToRelocations);
		for (int i=0;i<h.NumberOfRelocations;++i)
		{
			ssc << dump_relocation(rel0[i], cd.base + h.PointerToRawData, cd);
		}
	}
	ssc << "\n";
	return ssc.str();
}



std::string dbobj_com_descriptor_directory::get_display()
{
	std::stringstream ssc;

	IMAGE_COR20_HEADER * hdr = (IMAGE_COR20_HEADER*)this->get_start();
	cofpcdat & cd = this->get_cd();

	ssc << "  clr Header:\n\n";
	ssc << "        " << lpad(hex(hdr->cb), 8) << " cb\n";
	ssc << "        " << lpad(majmin(hdr->MajorRuntimeVersion, hdr->MinorRuntimeVersion), 8) << " runtime version\n";
	ssc << dump_ddinfo(&hdr->MetaData, "MetaData");
	ssc << "        " << lpad(hex(hdr->Flags), 8) << " flags\n";
	ssc << "        " << lpad(hex(hdr->EntryPointToken), 8) << " entry point";
	if (cd.test_rva(hdr->EntryPointToken))
	{
		ssc << " (" << lpad(hex(cd.imagebase + hdr->EntryPointToken), 8, '0') << ")";
		std::string alias;
		if (cd.test_symtable_alias(hdr->EntryPointToken, alias))
		{
			ssc << " " << alias;
		}
	}
	else
	{
		ssc << " token";
	}
	ssc << "\n";
	ssc << dump_ddinfo(&hdr->Resources, "Resources");
	ssc << dump_ddinfo(&hdr->StrongNameSignature, "StrongNameSignature");
	ssc << dump_ddinfo(&hdr->CodeManagerTable, "CodeManagerTable");
	ssc << dump_ddinfo(&hdr->VTableFixups, "VTableFixups");
	ssc << dump_ddinfo(&hdr->ExportAddressTableJumps, "ExportAddressTableJumps");
	ssc << dump_ddinfo(&hdr->ManagedNativeHeader, "ManagedNativeHeader");
	ssc << "\n\n";

	return ssc.str();
}

typedef struct FIXUP_TE_
{
	unsigned int offset : 12;
	unsigned int type : 4;
} FIXUP_TE;

typedef struct FIXUP_BLOCK_
{
	DWORD PageRVA;
	DWORD BlockSize;
	WORD  fixups[1];
} FIXUP_BLOCK;

std::string dbobj_reloc_directory::get_display()
{
	std::stringstream ssc;
	cofpcdat & cd = this->get_cd();
	image_section_helper & sh = m_sh;

	int size = this->get_size();
	byte * start = this->get_start();
	byte * pd = start;

	char * fixup_type[] = { "ABS             ",
							"HIGH            ",
							"LOW             ",
							"HIGHLOW         ",
							"HIGHADJ         ",
							"MIPS_JUMPADDR   ",
							"SECTION         ",
							"REL32           ",
							"MIPS_JUMPADDR16 ",
							"DIR64           ",
							"HIGHADJ         " };

	ssc << "BASE RELOCATIONS #" << hex(sh.section_id()) << "\n";
	do {
		FIXUP_BLOCK * pfixup = (FIXUP_BLOCK*)pd;
		ssc << lpad(hex(pfixup->PageRVA), 8) << " RVA, " << lpad(hex(pfixup->BlockSize), 8) << " SizeOfBlock\n";
		int nitems = (pfixup->BlockSize - 8) / 2;
		for (int i=0;i<nitems;++i)
		{
			FIXUP_TE * pf = (FIXUP_TE*)&pfixup->fixups[i];
			ssc << lpad(hex(pf->offset), 8) << "  " << fixup_type[pf->type] << "   ";
			switch (pf->type) 
			{
			case IMAGE_REL_BASED_ABSOLUTE:
				ssc << "        ";
				break;
			default: 
				{
					unsigned long value = *(DWORD*)cd.get_rva(pfixup->PageRVA + pf->offset); 
					ssc << lpad(hex(value), 8, '0');
					std::string alias;
					if (cd.test_symtable_alias(value - cd.imagebase, alias))
					{
						ssc << "  " << alias;
					}
				}
				break;
			}
			ssc << "\n";
		}
		nitems +=  1;  
		nitems &= ~1; 
		nitems +=  4;
		nitems *=  2;
		pd += nitems;
	} while (pd - start < size);
	ssc << "\n";
	return ssc.str();
}

std::string bound_import_str(byte * start, IMAGE_BOUND_IMPORT_DESCRIPTOR * bid)
{
	std::stringstream ss;
	ss << " to " << (char*)start + bid->OffsetModuleName;
	ss << " [" << lpad(hex(bid->TimeDateStamp), 8, bid->TimeDateStamp ? '0' : ' ') << "]";
	ss << stamptotime(bid->TimeDateStamp, true);
	return ss.str();
}

std::string dbobj_bound_import_directory::get_display()
{
	byte * start = this->get_start();
	IMAGE_BOUND_IMPORT_DESCRIPTOR * bid = (IMAGE_BOUND_IMPORT_DESCRIPTOR *)start;
	std::stringstream ssc;

	ssc << "  Header contains the following bound import information:\n";
	do 
	{
		if (!bid->OffsetModuleName) break;
		ssc << "    Bound" << bound_import_str(start, bid) << "\n";
		int nfr = bid->NumberOfModuleForwarderRefs;
		++bid;
		for (int i=0;i<nfr;++i)
		{
			ssc << "      Contained forwarders bound" << bound_import_str(start, bid) << "\n";
			++bid;
		}
	} while (true);
	ssc << "\n";

	return ssc.str();
}

std::string dbobj_tls_directory::get_display()
{
	std::stringstream ssc;
	image_section_helper & sh = m_sh;
	cofpcdat & cd = this->get_cd();

	IMAGE_TLS_DIRECTORY32 * tlsi = (IMAGE_TLS_DIRECTORY32*)this->get_start();

	ssc << "  Section contains the following TLS directory:\n\n";
	ssc << "            " << lpad(hex(tlsi->StartAddressOfRawData), 8, '0') << " Start of raw data\n";
	ssc << "            " << lpad(hex(tlsi->EndAddressOfRawData), 8, '0') << " End of raw data\n";
	ssc << "            " << lpad(hex(tlsi->AddressOfIndex), 8, '0') << " Address of index\n";
	ssc << "            " << lpad(hex(tlsi->AddressOfCallBacks), 8, '0') << " Address of callbacks\n";
	ssc << "            " << lpad(hex(tlsi->SizeOfZeroFill), 8) << " Size of zero fill\n";
	ssc << "            " << lpad(hex(tlsi->Characteristics), 8, '0') << " Characteristics\n";
	ssc << "\n";
	// address of a list of function pointers to be called when thread is inited, 0 terminated.
	if (tlsi->AddressOfCallBacks)
	{
		ssc << "    TLS Callbacks\n\n";
		ssc << "          Address\n";
		ssc << "          --------\n";
		DWORD * pcb = (DWORD*)sh.get_rva(tlsi->AddressOfCallBacks - cd.imagebase);
		do
		{
			ssc << "          " << lpad(hex(*pcb), 8, '0') << "\n";
		} while (*(pcb++));
		ssc << "\n";
	}

	return ssc.str();
}

std::string section_flags(unsigned long mask, int id, cofpcdat & cd)
{
	std::string pad("         ");
	std::stringstream ss;

	unsigned long i = 1;
	for (int x=0;x<20;++x, i<<=1)
	{
		unsigned long mm = mask & i;
		switch (mm)
		{
		case IMAGE_SCN_TYPE_NO_PAD:					ss << pad << "No Pad\n"; break;
		case IMAGE_SCN_CNT_CODE:					ss << pad << "Code\n"; break;
		case IMAGE_SCN_CNT_INITIALIZED_DATA:		ss << pad << "Initialized Data\n"; break;
		case IMAGE_SCN_CNT_UNINITIALIZED_DATA:		ss << pad << "Uninitialized Data\n"; break;
		case IMAGE_SCN_LNK_OTHER:					ss << pad << "Other\n"; break;
		case IMAGE_SCN_LNK_INFO:					ss << pad << "Info\n"; break;
		case IMAGE_SCN_LNK_REMOVE:					ss << pad << "Remove\n"; break;
		case IMAGE_SCN_LNK_COMDAT:					
			ss << pad << "COMDAT"; 
			{
				cofpcdat::iimapit i = cd.comdatrefs.find(id);
				if (i == cd.comdatrefs.end())
				{
					ss << " (no symbol)\n";
				}
				else
				{
					std::string name = symbol_name(cd.psym[i->second], cd.stringtable);
					std::string umname = unmangle(name);
					if (umname != "")
					{
						ss << "; sym= \"" << umname << "\" (" << name << ")\n";
					} 
					else
					{
						ss << "; sym= " << name << "\n";
					}
				}
			}
			break; // ?? SYMBOL INFO????
		case IMAGE_SCN_NO_DEFER_SPEC_EXC:			ss << pad << "No Defer Spec Exc\n"; break;
		case IMAGE_SCN_GPREL:						ss << pad << "GpRel|Mem FarData\n"; break;
		case IMAGE_SCN_MEM_PURGEABLE:				ss << pad << "Mem Purgeable|Mem 16bit\n"; break;
		case IMAGE_SCN_MEM_LOCKED:					ss << pad << "Mem Locked\n"; break;
		case IMAGE_SCN_MEM_PRELOAD:					ss << pad << "Mem Preload\n"; break;
		default:
			break;
		}
	}

	i = 0x01000000;
	for (int x=0;x<5;++x, i<<=1)
	{
		unsigned long mm = mask & i;
		switch (mm)
		{
		case IMAGE_SCN_LNK_NRELOC_OVFL:	ss << pad << "Contains extended relocations\n"; break;
		case IMAGE_SCN_MEM_DISCARDABLE:	ss << pad << "Discardable\n"; break;
		case IMAGE_SCN_MEM_NOT_CACHED:	ss << pad << "NonCacheable\n"; break;
		case IMAGE_SCN_MEM_NOT_PAGED:	ss << pad << "Not Paged\n"; break;
		case IMAGE_SCN_MEM_SHARED:		ss << pad << "Shared\n"; break;
		}
	}

	if (mask & IMAGE_SCN_ALIGN_MASK)
	{
		ss << pad;
		switch (mask & IMAGE_SCN_ALIGN_MASK)
		{
		case IMAGE_SCN_ALIGN_1BYTES:	ss << 1; break;
		case IMAGE_SCN_ALIGN_2BYTES:	ss << 2; break; 
		case IMAGE_SCN_ALIGN_4BYTES:	ss << 4; break;
		case IMAGE_SCN_ALIGN_8BYTES:	ss << 8; break;
		case IMAGE_SCN_ALIGN_16BYTES:	ss << 16; break;// default
		case IMAGE_SCN_ALIGN_32BYTES:	ss << 32; break;
		case IMAGE_SCN_ALIGN_64BYTES:	ss << 64; break;
		case IMAGE_SCN_ALIGN_128BYTES:	ss << 128; break;
		case IMAGE_SCN_ALIGN_256BYTES:	ss << 256; break;
		case IMAGE_SCN_ALIGN_512BYTES:	ss << 512; break;
		case IMAGE_SCN_ALIGN_1024BYTES: ss << 1024; break;
		case IMAGE_SCN_ALIGN_2048BYTES: ss << 2048; break;
		case IMAGE_SCN_ALIGN_4096BYTES: ss << 4096; break;
		case IMAGE_SCN_ALIGN_8192BYTES: ss << 8192; break;
			break;
		default:
			break;
		}
		ss << " byte align\n";
	}
	else if (cd.ident_type == IDENTIFY_LIB)
	{
		// only dumped for libraries ?
		ss << pad << "(no align specified)\n";
	}

	switch (mask & (IMAGE_SCN_MEM_EXECUTE|IMAGE_SCN_MEM_READ|IMAGE_SCN_MEM_WRITE))
	{
	case IMAGE_SCN_MEM_EXECUTE:	ss << pad << "Executable\n"; break;
	case IMAGE_SCN_MEM_READ:	ss << pad << "Read Only\n"; break; 
	case IMAGE_SCN_MEM_WRITE:	ss << pad << "Writeable\n"; break;
	case IMAGE_SCN_MEM_EXECUTE|IMAGE_SCN_MEM_READ:	ss << pad << "Execute Read\n"; break;
	case IMAGE_SCN_MEM_EXECUTE|IMAGE_SCN_MEM_WRITE:	ss << pad << "Execute Write\n"; break;
	case IMAGE_SCN_MEM_READ|IMAGE_SCN_MEM_WRITE:	ss << pad << "Read Write\n"; break;
	case IMAGE_SCN_MEM_READ|IMAGE_SCN_MEM_WRITE|IMAGE_SCN_MEM_EXECUTE:	ss << pad << "Execute Read Write\n"; break;
	default:
		break;
	}

	return ss.str();
}


std::string dbobj_section_header::get_display()
{
	std::stringstream ssc;

	image_section_helper & sh = this->m_s->get_sh();
	IMAGE_SECTION_HEADER & h = *(this->m_s->get_section_header());
	cofpcdat & cd = this->m_s->get_cd();

	ssc << "SECTION HEADER #" << hex(this->m_s->get_id()) << "\n";
	ssc << lpad(sh.name(), 8)	<< " name";
	if (sh.name()[0] == '/')
	{
		// boost::lexical_cast
		ssc << " (" <<  cd.stringtable_string(sh.name()) << ")";
	}
	ssc << "\n";
	if (h.Misc.VirtualSize)
	{
		ssc << lpad(hex(h.Misc.VirtualSize), 8)				<< " virtual size\n";
	}
	else
	{
		ssc << lpad(hex(h.Misc.PhysicalAddress), 8)			<< " physical address\n";
	}
	ssc << lpad(hex(h.VirtualAddress), 8)					<< " virtual address" << hex_range(cd.imagebase + h.VirtualAddress, h.Misc.VirtualSize) << "\n";
	ssc << lpad(hex(h.SizeOfRawData), 8)					<< " size of raw data\n";
	ssc << lpad(hex(h.PointerToRawData), 8)				<< " file pointer to raw data" << (h.SizeOfRawData ? hex_range(h.PointerToRawData, h.SizeOfRawData) : "") << "\n";
	ssc << lpad(hex(h.PointerToRelocations), 8)			<< " file pointer to relocation table\n";
	ssc << lpad(hex(h.PointerToLinenumbers), 8)			<< " file pointer to line numbers\n";
	ssc << lpad(hex(h.NumberOfRelocations), 8)			<< " number of relocations\n";
	ssc << lpad(hex(h.NumberOfLinenumbers), 8)			<< " number of line numbers\n";
	ssc << lpad(hex(h.Characteristics), 8)				<< " flags\n";
	if (h.Characteristics)
	{
		ssc << section_flags(h.Characteristics, this->m_s->get_id(), cd);
	}
	return ssc.str();	
}

std::string dbobj_section_data::get_display()
{
	image_section_helper & sh = this->m_s->get_sh();
	cofpcdat &  cd = this->m_s->get_cd();
	int id = this->m_s->get_id();
	IMAGE_SECTION_HEADER & h = *(this->m_s->get_section_header());

	std::stringstream ssc;

	unsigned long size = sh.dumpsize();
	ssc << "RAW DATA #" << hex(id) << "\n";
	ssc << hexdump(cd.base + h.PointerToRawData, size, cd.imagebase + h.VirtualAddress);
	ssc << "\n";

	return ssc.str();
}

std::string dbobj_section_linker_directives::get_display()
{
	cofpcdat & cd = this->m_s->get_cd();
	IMAGE_SECTION_HEADER & h = *(this->m_s->get_section_header());

	std::stringstream ssc;

	// This should be split up (at spaces)? into seperate lines to mirror dumpbin...
	ssc << "   Linker Directives\n";
	ssc << "   -----------------\n";
	byte * start = cd.base + h.PointerToRawData;
	unsigned int npadspace = 0;
	std::string spaces;
	for (;*start == 0x20 && npadspace < h.SizeOfRawData; ++start, ++npadspace, spaces += " ");
	
	std::string directives = str(start, h.SizeOfRawData - npadspace);
	std::vector<std::string> vdirectives;
	split(vdirectives, directives, " ");

	for (std::vector<std::string>::iterator vit = vdirectives.begin(); vit != vdirectives.end(); ++vit)
	{
		ssc << spaces << *vit << "\n";
	}
	ssc << "\n";

	return ssc.str();
}

std::string dbobj_coff_summary::get_display()
{
	std::stringstream ssc;
	ssc << "  Summary\n\n";
	for (std::vector<summary_item>::iterator vit = m_si.begin(); vit != m_si.end(); ++vit)
	{
		ssc << lpad(hex(vit->loc), 12) << " " << vit->name << "\n";
	}
	return ssc.str();
}

std::string dbobj_coff_section::get_display()
{
	dbobj_section_header dtmp(this);
	return dtmp.get_display();
}

std::string dbobj_anonymous_object::get_display()
{
	std::stringstream ss;
		
	ss << "File Type: ANONYMOUS OBJECT\n\n";
	ss << "ANON OBJECT HEADER VALUES\n";
	ss << lpad(hex(m_hdr->Version), 16) << " version\n";
	ss << lpad(hex(m_hdr->Machine), 16) << " machine (" << machine_type(m_hdr->Machine) << ")\n";
	ss << lpad(hex(m_hdr->TimeDateStamp), 16) << " time date stamp" <<  stamptotime(m_hdr->TimeDateStamp) << "\n";
	ss << lpad("", 16) << " ClassID: ";
	{
		USES_CONVERSION;
		OLECHAR tmp[128];
		StringFromGUID2(m_hdr->ClassID, tmp, sizeof(tmp));
		ss << OLE2A(tmp) << "\n";
	}
	// simulate a bug in dumpbin
	ss << lpad(hex(0xffff & m_hdr->SizeOfData), 16) << " size\n";
	ss << lpad(hex(0), 16) << " flags\n";
	return ss.str();
}


std::string dbobj_coff_import_obj::get_display()
{
	std::stringstream ssc;
	byte * data = m_base;

	IMPORT_OBJECT_HEADER & h = *(IMPORT_OBJECT_HEADER *)data;

	data += sizeof(IMPORT_OBJECT_HEADER);
	std::string name_imp = (char*)data;
	std::string undec_name_imp = name_imp;
	data += name_imp.length() + 1;
	std::string name_dll = (char*)data;

	ssc << "  Version      : " << hex(h.Version) << "\n";
	ssc << "  Machine      : " << hex(h.Machine) << " (" << machine_type(h.Machine) << ")\n";
	ssc << "  TimeDateStamp: " << hex(h.TimeDateStamp) << stamptotime(h.TimeDateStamp) << "\n";
	ssc << "  SizeOfData   : " << lpad(hex(h.SizeOfData),8,'0') << "\n";
	ssc << "  DLL name     : " << name_dll << "\n";
	std::string undec = unmangle(name_imp);
	if (undec != "")
	{
		std::stringstream ss;
		ss << " (" << undec << ")";
		name_imp += ss.str();
	}
	ssc << "  Symbol name  : " << name_imp << "\n";
	ssc << "  Type         : " << import_object_header_type(h.Type) << "\n";
	ssc << "  Name type    : " << import_object_header_name_type(h.NameType) << "\n";
	ssc << "  Hint         : " << h.Hint << "\n";
	ssc << "  Name         : " << undec_name_imp << "\n";
	ssc << "\n";

	return ssc.str();
}

std::string dbobj_image_archive_member_header::get_display()
{
	std::stringstream ssc;

	ssc << "Archive member name at " << hex(m_offset) << ": ";

	IMAGE_ARCHIVE_MEMBER_HEADER & h = *m_ph;

	long td = atoi(str(h.Date,12).c_str());
	ssc <<    rpad(str(h.Name,16),16)         << this->m_longname << "\n";
	ssc <<	lpad(hex(td),8)					<< " time/date" << stamptotime(td) << "\n";
	ssc <<	lpad(hex(str(h.UserID,4)), 8)	<< " uid\n";
	ssc <<	lpad(hex(str(h.GroupID,4)), 8)	<< " gid\n";
	ssc <<	lpad(str(h.Mode,8), 8)			<< " mode\n";
	ssc <<	lpad(hex(str(h.Size,10)), 8)	<< " size\n";
	
	if (h.EndHeader[0] == 0x60 && h.EndHeader[1] == 0x0A) 
	{
		ssc << "correct header end\n\n";
	}
	else
	{
		ssc << "incorrect header end\n\n";
		throw std::runtime_error("incorrect header end");
	}

	return ssc.str();
}

std::string dbobj_image_archive_member_header::get_name() const
{
	std::stringstream ss;
	ss << "Archive member #" << m_idx;
	switch (mem_type)
	{
	case dbo_iamh_linker_member:
		ss << " (linker member)";
		break;
	case dbo_iamh_longnames_member:
		ss << " (longnames member)";
		break;
	default:
		break;
	}
	return ss.str();
}

std::string dbobj_iamh_symbols::get_display()
{
	byte * data2 = m_data;

	std::stringstream ssc;

	long nsym = be2le(*(long*)data2);
	data2 += sizeof(long);
	
	ssc << lpad(str(nsym), 5) << " public symbols\n\n";

	byte * pnames = data2 + (nsym*sizeof(long));
	unsigned long * poffs  = (unsigned long*)data2;
	for (int i=0;i<nsym;++i)
	{
		long offset = be2le(*poffs);
		++poffs;

		int len = (int)strlen((const char*)pnames);
		std::string name((char*)pnames, len);
		pnames += len+1;

		ssc << lpad(hex(offset), 9) << " " << name << "\n";
	}
	ssc << "\n";

	return ssc.str();
}

std::string dbobj_iamh_offsets::get_display()
{

	byte * data2 = m_data;
	std::stringstream ssc;

	long noffs = (*(long*)data2);
	data2 += sizeof(long);

	ssc << lpad(str(noffs), 5) << " offsets\n\n";

	for (int i=0;i<noffs;++i)
	{
		long offset = (*(long*)data2);
		data2 += sizeof(long);			
		ssc << lpad(hex(i+1), 9) << lpad(hex(offset), 9) << "\n";
	}
	ssc << "\n";

	long nsym = (*(long*)data2);
	data2 += sizeof(long);
	ssc << lpad(str(nsym), 5) << " public symbols\n\n";

	byte * pnames = data2 + (nsym * sizeof(short));
	for (int i=0;i<nsym;++i)
	{
		unsigned short hint = *(unsigned short *)data2;
		data2 += sizeof(unsigned short);
		int len = (int)strlen((const char*)pnames);
		std::string name((char*)pnames, len);
		pnames += len+1;
		ssc << lpad(hex(hint), 9) << " " << name << "\n";
	}
	ssc << "\n";

	return ssc.str();
}

std::string dbobj_coff_exports::get_display()
{
	std::stringstream ssc;

	ssc << "     Exports\n\n";
	ssc << "       ordinal    name\n\n";
	for (std::vector<coff_export_symbol>::const_iterator vit = m_cev.begin(); vit != m_cev.end(); ++vit)
	{
		ssc << "                 " << " ";
		ssc << vit->symbol << "\n";
	}

	return ssc.str();
}

std::string dbobj_coff_symbol_table::get_display()
{
	std::stringstream ssc;

	cofpcdat & cd = m_cd;

	int nsym = cd.nsym;
	IMAGE_SYMBOL * psym = cd.psym;

	if (!nsym || !psym) return "";
	ssc << "COFF SYMBOL TABLE\n";

	for (int i = 0; i < nsym; ++i)
	{
		ssc << print_symbol(psym[i], cd.stringtable, i);
		if (psym[i].NumberOfAuxSymbols)
		{
			IMAGE_AUX_SYMBOL * paux = (IMAGE_AUX_SYMBOL *)psym+i+1;
			switch (psym[i].StorageClass)
			{
			case IMAGE_SYM_CLASS_STATIC:
				if (psym[i].Type == 32)
				{
					ssc << dump_aux_externals(psym[i], paux, psym[i].NumberOfAuxSymbols);
				}
				else
				{
					ssc << dump_aux_statics(psym[i], paux, psym[i].NumberOfAuxSymbols);
				}
				break;
			case IMAGE_SYM_CLASS_WEAK_EXTERNAL:
				ssc << dump_aux_weak_externals(psym[i], paux, psym[i].NumberOfAuxSymbols);
				break;
			case IMAGE_SYM_CLASS_FILE:
				ssc << dump_aux_file(psym[i], paux, psym[i].NumberOfAuxSymbols, cd.stringtable);
				break;
			case IMAGE_SYM_CLASS_EXTERNAL:
				ssc << dump_aux_externals(psym[i], paux, psym[i].NumberOfAuxSymbols);
				break;
			case IMAGE_SYM_CLASS_FUNCTION:
				{
				std::string name = symbol_name(psym[i], cd.stringtable);
				ssc << dump_aux_function(psym[i], paux, psym[i].NumberOfAuxSymbols, name);
				}
				break;
			default:
				throw std::runtime_error("Unknown aux symbols");
			}
			i += psym[i].NumberOfAuxSymbols;
		}
	}
	ssc << "\n";
	
	return ssc.str();
}